@@ -27,10 +27,6 @@
 #include "gui/DatabaseTabWidget.h"
 #include "gui/DatabaseWidget.h"
 
-#include <QDBusConnection>
-#include <QDBusServiceWatcher>
-#include <QDebug>
-
 namespace
 {
     constexpr auto DEFAULT_ALIAS = "default";
@@ -38,44 +34,38 @@ namespace
 
 namespace FdoSecrets
 {
+    QSharedPointer<Service>
+    Service::Create(FdoSecretsPlugin* plugin, QPointer<DatabaseTabWidget> dbTabs, QSharedPointer<DBusMgr> dbus)
+    {
+        QSharedPointer<Service> res{new Service(plugin, std::move(dbTabs), std::move(dbus))};
+        if (!res->initialize()) {
+            return {};
+        }
+        return res;
+    }
 
     Service::Service(FdoSecretsPlugin* plugin,
-                     QPointer<DatabaseTabWidget> dbTabs) // clazy: exclude=ctor-missing-parent-argument
-        : DBusObject(nullptr)
+                     QPointer<DatabaseTabWidget> dbTabs,
+                     QSharedPointer<DBusMgr> dbus) // clazy: exclude=ctor-missing-parent-argument
+        : DBusObject(std::move(dbus))
         , m_plugin(plugin)
         , m_databases(std::move(dbTabs))
-        , m_insdieEnsureDefaultAlias(false)
-        , m_serviceWatcher(nullptr)
+        , m_insideEnsureDefaultAlias(false)
     {
-        connect(
-            m_databases, &DatabaseTabWidget::databaseUnlockDialogFinished, this, &Service::doneUnlockDatabaseInDialog);
+        connect(m_databases,
+                &DatabaseTabWidget::databaseUnlockDialogFinished,
+                this,
+                &Service::onDatabaseUnlockDialogFinished);
     }
 
-    Service::~Service()
-    {
-        QDBusConnection::sessionBus().unregisterService(QStringLiteral(DBUS_SERVICE_SECRET));
-    }
+    Service::~Service() = default;
 
     bool Service::initialize()
     {
-        if (!QDBusConnection::sessionBus().registerService(QStringLiteral(DBUS_SERVICE_SECRET))) {
-            qDebug() << "Another secret service is running";
-            emit error(tr("Failed to register DBus service at %1: another secret service is running.")
-                           .arg(QLatin1Literal(DBUS_SERVICE_SECRET)));
+        if (!dbus()->registerObject(this)) {
             return false;
         }
 
-        registerWithPath(QStringLiteral(DBUS_PATH_SECRETS), new ServiceAdaptor(this));
-
-        // Connect to service unregistered signal
-        m_serviceWatcher.reset(new QDBusServiceWatcher());
-        connect(m_serviceWatcher.data(),
-                &QDBusServiceWatcher::serviceUnregistered,
-                this,
-                &Service::dbusServiceUnregistered);
-
-        m_serviceWatcher->setConnection(QDBusConnection::sessionBus());
-
         // Add existing database tabs
         for (int idx = 0; idx != m_databases->count(); ++idx) {
             auto dbWidget = m_databases->databaseWidgetFromIndex(idx);
@@ -89,7 +79,7 @@ namespace FdoSecrets
         });
 
         // make default alias track current activated database
-        connect(m_databases.data(), &DatabaseTabWidget::activateDatabaseChanged, this, &Service::ensureDefaultAlias);
+        connect(m_databases.data(), &DatabaseTabWidget::activeDatabaseChanged, this, &Service::ensureDefaultAlias);
 
         return true;
     }
@@ -98,55 +88,64 @@ namespace FdoSecrets
     {
         // The Collection will monitor the database's exposed group.
         // When the Collection finds that no exposed group, it will delete itself.
-        // Thus the service also needs to monitor it and recreate the collection if the user changes
+        // Thus, the service also needs to monitor it and recreate the collection if the user changes
         // from no exposed to exposed something.
+        connect(dbWidget, &DatabaseWidget::databaseReplaced, this, [this, dbWidget]() {
+            monitorDatabaseExposedGroup(dbWidget);
+        });
         if (!dbWidget->isLocked()) {
             monitorDatabaseExposedGroup(dbWidget);
         }
-        connect(dbWidget, &DatabaseWidget::databaseUnlocked, this, [this, dbWidget]() {
-            monitorDatabaseExposedGroup(dbWidget);
-        });
 
-        auto coll = new Collection(this, dbWidget);
-        // Creation may fail if the database is not exposed.
-        // This is okay, because we monitor the expose settings above
-        if (!coll->isValid()) {
-            coll->deleteLater();
+        auto coll = Collection::Create(this, dbWidget);
+        if (!coll) {
             return;
         }
 
         m_collections << coll;
         m_dbToCollection[dbWidget] = coll;
 
-        // handle alias
+        // keep record of the collection existence
+        connect(coll, &Collection::collectionAboutToDelete, this, [this, coll]() {
+            m_collections.removeAll(coll);
+            m_dbToCollection.remove(coll->backend());
+        });
+
+        // keep record of alias
         connect(coll, &Collection::aliasAboutToAdd, this, &Service::onCollectionAliasAboutToAdd);
         connect(coll, &Collection::aliasAdded, this, &Service::onCollectionAliasAdded);
         connect(coll, &Collection::aliasRemoved, this, &Service::onCollectionAliasRemoved);
 
-        ensureDefaultAlias();
-
         // Forward delete signal, we have to rely on filepath to identify the database being closed,
         // but we can not access m_backend safely because during the databaseClosed signal,
         // m_backend may already be reset to nullptr
         // We want to remove the collection object from dbus as early as possible, to avoid
         // race conditions when deleteLater was called on the m_backend, but not delivered yet,
-        // and new method calls from dbus occurred. Therefore we can't rely on the destroyed
+        // and new method calls from dbus occurred. Therefore, we can't rely on the destroyed
         // signal on m_backend.
         // bind to coll lifespan
         connect(m_databases.data(), &DatabaseTabWidget::databaseClosed, coll, [coll](const QString& filePath) {
             if (filePath == coll->backendFilePath()) {
-                coll->doDelete();
+                coll->removeFromDBus();
             }
         });
 
-        // relay signals
-        connect(coll, &Collection::collectionChanged, this, [this, coll]() { emit collectionChanged(coll); });
-        connect(coll, &Collection::collectionAboutToDelete, this, [this, coll]() {
-            m_collections.removeAll(coll);
-            m_dbToCollection.remove(coll->backend());
-            emit collectionDeleted(coll);
-        });
+        // actual load, must after updates to m_collections, because reloading may trigger
+        // another onDatabaseTabOpen, and m_collections will be used to prevent recursion.
+        if (!coll->reloadBackend()) {
+            // error in dbus
+            return;
+        }
+        if (!coll->backend()) {
+            // no exposed group on this db
+            return;
+        }
 
+        ensureDefaultAlias();
+
+        // only start relay signals when the collection is fully setup
+        connect(coll, &Collection::collectionChanged, this, [this, coll]() { emit collectionChanged(coll); });
+        connect(coll, &Collection::collectionAboutToDelete, this, [this, coll]() { emit collectionDeleted(coll); });
         if (emitSignal) {
             emit collectionCreated(coll);
         }
@@ -155,21 +154,20 @@ namespace FdoSecrets
     void Service::monitorDatabaseExposedGroup(DatabaseWidget* dbWidget)
     {
         Q_ASSERT(dbWidget);
-        connect(
-            dbWidget->database()->metadata()->customData(), &CustomData::customDataModified, this, [this, dbWidget]() {
-                if (!FdoSecrets::settings()->exposedGroup(dbWidget->database()).isNull() && !findCollection(dbWidget)) {
-                    onDatabaseTabOpened(dbWidget, true);
-                }
-            });
+        connect(dbWidget->database()->metadata()->customData(), &CustomData::modified, this, [this, dbWidget]() {
+            if (!FdoSecrets::settings()->exposedGroup(dbWidget->database()).isNull() && !findCollection(dbWidget)) {
+                onDatabaseTabOpened(dbWidget, true);
+            }
+        });
     }
 
     void Service::ensureDefaultAlias()
     {
-        if (m_insdieEnsureDefaultAlias) {
+        if (m_insideEnsureDefaultAlias) {
             return;
         }
 
-        m_insdieEnsureDefaultAlias = true;
+        m_insideEnsureDefaultAlias = true;
 
         auto coll = findCollection(m_databases->currentDatabaseWidget());
         if (coll) {
@@ -177,150 +175,161 @@ namespace FdoSecrets
             coll->addAlias(DEFAULT_ALIAS).okOrDie();
         }
 
-        m_insdieEnsureDefaultAlias = false;
+        m_insideEnsureDefaultAlias = false;
     }
 
-    void Service::dbusServiceUnregistered(const QString& service)
+    DBusResult Service::collections(QList<Collection*>& collections) const
     {
-        Q_ASSERT(m_serviceWatcher);
-
-        auto removed = m_serviceWatcher->removeWatchedService(service);
-        Q_UNUSED(removed);
-        Q_ASSERT(removed);
-
-        Session::CleanupNegotiation(service);
-        auto sess = m_peerToSession.value(service, nullptr);
-        if (sess) {
-            sess->close().okOrDie();
-        }
+        collections = m_collections;
+        return {};
     }
 
-    DBusReturn<const QList<Collection*>> Service::collections() const
+    DBusResult Service::openSession(const DBusClientPtr& client,
+                                    const QString& algorithm,
+                                    const QVariant& input,
+                                    QVariant& output,
+                                    Session*& result)
     {
-        return m_collections;
-    }
-
-    DBusReturn<QVariant> Service::openSession(const QString& algorithm, const QVariant& input, Session*& result)
-    {
-        QVariant output;
-        bool incomplete = false;
-        auto peer = callingPeer();
-
-        // watch for service unregister to cleanup
-        Q_ASSERT(m_serviceWatcher);
-        m_serviceWatcher->addWatchedService(peer);
-
         // negotiate cipher
-        auto ciphers = Session::CreateCiphers(peer, algorithm, input, output, incomplete);
+        bool incomplete = false;
+        auto ciphers = client->negotiateCipher(algorithm, input, output, incomplete);
         if (incomplete) {
             result = nullptr;
-            return output;
+            return {};
         }
         if (!ciphers) {
-            return DBusReturn<>::Error(QDBusError::NotSupported);
+            return QDBusError::NotSupported;
         }
-        result = new Session(std::move(ciphers), callingPeerName(), this);
 
-        m_sessions.append(result);
-        m_peerToSession[peer] = result;
-        connect(result, &Session::aboutToClose, this, [this, peer, result]() {
-            emit sessionClosed(result);
-            m_sessions.removeAll(result);
-            m_peerToSession.remove(peer);
+        // create session using the negotiated cipher
+        result = Session::Create(std::move(ciphers), client->name(), this);
+        if (!result) {
+            return QDBusError::InternalError;
+        }
+
+        // remove session when the client disconnects
+        connect(dbus().data(), &DBusMgr::clientDisconnected, result, [result, client](const DBusClientPtr& toRemove) {
+            if (toRemove == client) {
+                result->close().okOrDie();
+            }
         });
-        emit sessionOpened(result);
 
-        return output;
+        // keep a list of sessions
+        m_sessions.append(result);
+        connect(result, &Session::aboutToClose, this, [this, result]() { m_sessions.removeAll(result); });
+
+        return {};
     }
 
-    DBusReturn<Collection*>
-    Service::createCollection(const QVariantMap& properties, const QString& alias, PromptBase*& prompt)
+    DBusResult Service::createCollection(const QVariantMap& properties,
+                                         const QString& alias,
+                                         Collection*& collection,
+                                         PromptBase*& prompt)
     {
         prompt = nullptr;
 
         // return existing collection if alias is non-empty and exists.
-        auto collection = findCollection(alias);
+        collection = findCollection(alias);
         if (!collection) {
-            auto cp = new CreateCollectionPrompt(this);
-            prompt = cp;
-
-            // collection will be created when the prompt complets.
-            // once it's done, we set additional properties on the collection
-            connect(cp, &CreateCollectionPrompt::collectionCreated, cp, [alias, properties](Collection* coll) {
-                coll->setProperties(properties).okOrDie();
-                if (!alias.isEmpty()) {
-                    coll->addAlias(alias).okOrDie();
-                }
-            });
+            prompt = PromptBase::Create<CreateCollectionPrompt>(this, properties, alias);
+            if (!prompt) {
+                return QDBusError::InternalError;
+            }
         }
-        return collection;
+        return {};
     }
 
-    DBusReturn<const QList<Item*>> Service::searchItems(const StringStringMap& attributes, QList<Item*>& locked)
+    DBusResult Service::searchItems(const DBusClientPtr& client,
+                                    const StringStringMap& attributes,
+                                    QList<Item*>& unlocked,
+                                    QList<Item*>& locked) const
     {
-        auto ret = collections();
-        if (ret.isError()) {
+        QList<Collection*> colls;
+        auto ret = collections(colls);
+        if (ret.err()) {
             return ret;
         }
 
-        QList<Item*> unlocked;
-        for (const auto& coll : ret.value()) {
-            auto items = coll->searchItems(attributes);
-            if (items.isError()) {
-                return items;
+        for (const auto& coll : asConst(colls)) {
+            QList<Item*> items;
+            ret = coll->searchItems(client, attributes, items);
+            if (ret.err()) {
+                return ret;
             }
-            auto l = coll->locked();
-            if (l.isError()) {
-                return l;
-            }
-            if (l.value()) {
-                locked.append(items.value());
-            } else {
-                unlocked.append(items.value());
-            }
-        }
-        return unlocked;
-    }
-
-    DBusReturn<const QList<DBusObject*>> Service::unlock(const QList<DBusObject*>& objects, PromptBase*& prompt)
-    {
-        QSet<Collection*> needUnlock;
-        needUnlock.reserve(objects.size());
-        for (const auto& obj : asConst(objects)) {
-            auto coll = qobject_cast<Collection*>(obj);
-            if (coll) {
-                needUnlock << coll;
-            } else {
-                auto item = qobject_cast<Item*>(obj);
-                if (!item) {
-                    continue;
+            // item locked state already covers its collection's locked state
+            for (const auto& item : asConst(items)) {
+                bool l;
+                ret = item->locked(client, l);
+                if (ret.err()) {
+                    return ret;
+                }
+                if (l) {
+                    locked.append(item);
+                } else {
+                    unlocked.append(item);
                 }
-                // we lock the whole collection for item
-                needUnlock << item->collection();
             }
         }
-
-        // return anything already unlocked
-        QList<DBusObject*> unlocked;
-        QList<Collection*> toUnlock;
-        for (const auto& coll : asConst(needUnlock)) {
-            auto l = coll->locked();
-            if (l.isError()) {
-                return l;
-            }
-            if (!l.value()) {
-                unlocked << coll;
-            } else {
-                toUnlock << coll;
-            }
-        }
-        if (!toUnlock.isEmpty()) {
-            prompt = new UnlockCollectionsPrompt(this, toUnlock);
-        }
-        return unlocked;
+        return {};
     }
 
-    DBusReturn<const QList<DBusObject*>> Service::lock(const QList<DBusObject*>& objects, PromptBase*& prompt)
+    DBusResult Service::unlock(const DBusClientPtr& client,
+                               const QList<DBusObject*>& objects,
+                               QList<DBusObject*>& unlocked,
+                               PromptBase*& prompt)
+    {
+        QSet<Collection*> collectionsToUnlock;
+        QSet<Item*> itemsToUnlock;
+        collectionsToUnlock.reserve(objects.size());
+        itemsToUnlock.reserve(objects.size());
+
+        for (const auto& obj : asConst(objects)) {
+            // the object is either an item or a collection
+            auto item = qobject_cast<Item*>(obj);
+            auto coll = item ? item->collection() : qobject_cast<Collection*>(obj);
+            // either way there should be a collection
+            if (!coll) {
+                continue;
+            }
+
+            bool collLocked{false}, itemLocked{false};
+            // if the collection needs unlock
+            auto ret = coll->locked(collLocked);
+            if (ret.err()) {
+                return ret;
+            }
+            if (collLocked) {
+                collectionsToUnlock << coll;
+            }
+
+            if (item) {
+                // item may also need unlock
+                ret = item->locked(client, itemLocked);
+                if (ret.err()) {
+                    return ret;
+                }
+                if (itemLocked) {
+                    itemsToUnlock << item;
+                }
+            }
+
+            // both collection and item are not locked
+            if (!collLocked && !itemLocked) {
+                unlocked << obj;
+            }
+        }
+
+        prompt = nullptr;
+        if (!collectionsToUnlock.isEmpty() || !itemsToUnlock.isEmpty()) {
+            prompt = PromptBase::Create<UnlockPrompt>(this, collectionsToUnlock, itemsToUnlock);
+            if (!prompt) {
+                return QDBusError::InternalError;
+            }
+        }
+        return {};
+    }
+
+    DBusResult Service::lock(const QList<DBusObject*>& objects, QList<DBusObject*>& locked, PromptBase*& prompt)
     {
         QSet<Collection*> needLock;
         needLock.reserve(objects.size());
@@ -339,60 +348,62 @@ namespace FdoSecrets
         }
 
         // return anything already locked
-        QList<DBusObject*> locked;
         QList<Collection*> toLock;
         for (const auto& coll : asConst(needLock)) {
-            auto l = coll->locked();
-            if (l.isError()) {
-                return l;
+            bool l;
+            auto ret = coll->locked(l);
+            if (ret.err()) {
+                return ret;
             }
-            if (l.value()) {
+            if (l) {
                 locked << coll;
             } else {
                 toLock << coll;
             }
         }
         if (!toLock.isEmpty()) {
-            prompt = new LockCollectionsPrompt(this, toLock);
+            prompt = PromptBase::Create<LockCollectionsPrompt>(this, toLock);
+            if (!prompt) {
+                return QDBusError::InternalError;
+            }
         }
-        return locked;
+        return {};
     }
 
-    DBusReturn<const QHash<Item*, SecretStruct>> Service::getSecrets(const QList<Item*>& items, Session* session)
+    DBusResult Service::getSecrets(const DBusClientPtr& client,
+                                   const QList<Item*>& items,
+                                   Session* session,
+                                   ItemSecretMap& secrets) const
     {
         if (!session) {
-            return DBusReturn<>::Error(QStringLiteral(DBUS_ERROR_SECRET_NO_SESSION));
+            return DBusResult(DBUS_ERROR_SECRET_NO_SESSION);
         }
 
-        QHash<Item*, SecretStruct> res;
-
         for (const auto& item : asConst(items)) {
-            auto ret = item->getSecret(session);
-            if (ret.isError()) {
+            auto ret = item->getSecretNoNotification(client, session, secrets[item]);
+            if (ret.err()) {
                 return ret;
             }
-            res[item] = std::move(ret).value();
         }
-        if (calledFromDBus()) {
-            plugin()->emitRequestShowNotification(
-                tr(R"(%n Entry(s) was used by %1)", "%1 is the name of an application", res.size())
-                    .arg(callingPeerName()));
-        }
-        return res;
+        plugin()->emitRequestShowNotification(
+            tr(R"(%n Entry(s) was used by %1)", "%1 is the name of an application", secrets.size())
+                .arg(client->name()));
+        return {};
     }
 
-    DBusReturn<Collection*> Service::readAlias(const QString& name)
+    DBusResult Service::readAlias(const QString& name, Collection*& collection) const
     {
-        return findCollection(name);
+        collection = findCollection(name);
+        return {};
     }
 
-    DBusReturn<void> Service::setAlias(const QString& name, Collection* collection)
+    DBusResult Service::setAlias(const QString& name, Collection* collection)
     {
         if (!collection) {
             // remove alias name from its collection
             collection = findCollection(name);
             if (!collection) {
-                return DBusReturn<>::Error(QStringLiteral(DBUS_ERROR_SECRET_NO_SUCH_OBJECT));
+                return DBusResult(DBUS_ERROR_SECRET_NO_SUCH_OBJECT);
             }
             return collection->removeAlias(name);
         }
@@ -444,7 +455,7 @@ namespace FdoSecrets
         return m_dbToCollection.value(db, nullptr);
     }
 
-    const QList<Session*> Service::sessions() const
+    QList<Session*> Service::sessions() const
     {
         return m_sessions;
     }
@@ -472,22 +483,57 @@ namespace FdoSecrets
         return collection;
     }
 
-    void Service::doSwitchToChangeDatabaseSettings(DatabaseWidget* dbWidget)
+    void Service::doSwitchToDatabaseSettings(DatabaseWidget* dbWidget)
     {
         if (dbWidget->isLocked()) {
             return;
         }
         // switch selected to current
         m_databases->setCurrentWidget(dbWidget);
-        m_databases->changeDatabaseSettings();
+        m_databases->showDatabaseSettings();
 
         // open settings (switch from app settings to m_dbTabs)
         m_plugin->emitRequestSwitchToDatabases();
     }
 
+    bool Service::doLockDatabase(DatabaseWidget* dbWidget)
+    {
+        // return immediately if the db is already unlocked
+        if (dbWidget && dbWidget->isLocked()) {
+            return true;
+        }
+
+        // mark the db as being unlocked to prevent multiple dialogs for the same db
+        if (m_lockingDb.contains(dbWidget)) {
+            return true;
+        }
+        m_lockingDb.insert(dbWidget);
+        auto ret = dbWidget->lock();
+        m_lockingDb.remove(dbWidget);
+
+        return ret;
+    }
+
     void Service::doUnlockDatabaseInDialog(DatabaseWidget* dbWidget)
     {
+        // return immediately if the db is already unlocked
+        if (dbWidget && !dbWidget->isLocked()) {
+            emit doneUnlockDatabaseInDialog(true, dbWidget);
+            return;
+        }
+
+        // mark the db as being unlocked to prevent multiple dialogs for the same db
+        if (m_unlockingDb.contains(dbWidget)) {
+            return;
+        }
+        m_unlockingDb.insert(dbWidget);
+
         m_databases->unlockDatabaseInDialog(dbWidget, DatabaseOpenDialog::Intent::None);
     }
 
+    void Service::onDatabaseUnlockDialogFinished(bool accepted, DatabaseWidget* dbWidget)
+    {
+        m_unlockingDb.remove(dbWidget);
+        emit doneUnlockDatabaseInDialog(accepted, dbWidget);
+    }
 } // namespace FdoSecrets
