@@ -237,6 +237,7 @@ void DatabaseOpenWidget::openDatabase()
     m_blockQuickUnlock = false;
 
     setUserInteractionLock(true);
+    m_ui->editPassword->setShowPassword(false);
     m_ui->messageWidget->hide();
     QCoreApplication::processEvents();
 
@@ -455,7 +456,7 @@ void DatabaseOpenWidget::pollHardwareKey()
     m_ui->hardwareKeyProgress->setVisible(true);
     m_pollingHardwareKey = true;
 
-    YubiKey::instance()->findValidKeys();
+    YubiKey::instance()->findValidKeysAsync();
 }
 
 void DatabaseOpenWidget::hardwareKeyResponse(bool found)
@@ -528,8 +529,16 @@ bool DatabaseOpenWidget::isOnQuickUnlockScreen()
     return m_ui->centralStack->currentIndex() == 1;
 }
 
+/**
+ * Reset installed quick unlock secrets.
+ *
+ * It's safe to call this method even if quick unlock is unavailable.
+ */
 void DatabaseOpenWidget::resetQuickUnlock()
 {
+    if (!isQuickUnlockAvailable()) {
+        return;
+    }
 #if defined(Q_CC_MSVC)
     getWindowsHello()->reset(m_filename);
 #elif defined(Q_OS_MACOS)
